# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger <matthias@bilger.info>
import logging
import pathlib
from collections import defaultdict

import jinja2
import markdown
import yaml
from aiohttp import web
from aiohttp_jinja2 import render_template
from aiohttp_jinja2 import setup as jinja2_setup
from mutenix.config import Config
from mutenix.version import MAJOR
from mutenix.version import MINOR
from mutenix.version import PATCH

HOST = "127.0.0.1"
PORT = 12909

_logger = logging.getLogger(__name__)


class WebServer:
    """A general web server class for serving web pages and static content."""

    icons: list[dict[str, str]] = [
        {
            "src": "/favicon/32",
            "sizes": "32x32",
            "type": "image/png",
        },
        {
            "src": "/favicon/64",
            "sizes": "64x64",
            "type": "image/png",
        },
        {
            "src": "/favicon/16",
            "sizes": "16x16",
            "type": "image/png",
        },
        {
            "src": "/favicon/apple_touch",
            "sizes": "180x180",
            "type": "image/png",
        },
    ]

    def __init__(self, host: str = HOST, port: int = PORT):
        self._hardware = ""
        self._version = ""
        self.host = host
        self.port = port
        self._running = False
        self.app = web.Application()
        self.app.router.add_static(
            "/static/",
            path=str(pathlib.Path(__file__).parent / "static"),
            name="static",
        )
        self.app.router.add_route("GET", "/favicon/{filename}", self.favicon)
        self.app.router.add_route("GET", "/favicon.svg", self.favicon_svg)
        self.app.router.add_route("GET", "/favicon.ico", self.favicon_ico)
        self.app.router.add_route("GET", "/site.webmanifest", self.serve_manifest)
        self.app.router.add_route("GET", "/images/{name}", self.serve_image)
        self.app.router.add_route("GET", "/help", self.help)
        self.app.router.add_route("GET", "/about", self.about)
        self.app.router.add_route("GET", "/config", self.config)
        self.app.router.add_route("GET", "/device", self.device)
        self.app.add_routes(
            [
                web.get("/", self.index),
                web.get("/popup", self.popup),
            ],
        )
        jinja2_setup(self.app, loader=jinja2.PackageLoader("mutenix", "templates"))

    def _render_template(self, template_name, request, context):
        context["mutenix_version"] = f"{MAJOR}.{MINOR}.{PATCH}"
        return render_template(template_name, request, context)

    def update_config(self, config):
        self._config: Config = config

    async def index(self, request: web.Request):
        return self._render_template("index.html", request, {})

    async def popup(self, request: web.Request):
        return self._render_template("popup.html", request, {})

    async def serve_image(self, request: web.Request):
        name = request.match_info["name"]
        image_path = pathlib.Path(__file__).parent / "assets" / f"{name}"
        return web.FileResponse(image_path)

    async def favicon(self, request: web.Request):
        filename = request.match_info["filename"]
        for icon in self.icons:
            if icon["src"].endswith(filename):
                icon_path = (
                    pathlib.Path(__file__).parent
                    / "assets"
                    / f"icon_active_{filename}.png"
                )
                break
        else:
            raise web.HTTPNotFound()
        return web.FileResponse(icon_path)

    async def favicon_ico(self, request: web.Request):  # pragma: no cover
        return web.FileResponse(
            pathlib.Path(__file__).parent / "assets" / "mutenix.ico",
        )

    async def favicon_svg(self, request: web.Request):
        return web.FileResponse(
            pathlib.Path(__file__).parent
            / "assets"
            / "mutenix_logo_finalicon_active.svg",
        )

    async def serve_manifest(self, request: web.Request):
        manifest = {
            "name": "Mutenix Virtual Macropad",
            "short_name": "Mutenix",
            "icons": self.icons,
            "start_url": "/",
            "display": "standalone",
        }
        return web.json_response(manifest)

    async def help(self, request: web.Request):
        return self._render_template("help.html", request, {})

    async def about(self, request: web.Request):
        readme_path = pathlib.Path(__file__).parent / "README.md"
        license_path = pathlib.Path(__file__).parent / "LICENSE"
        try:
            with open(readme_path) as f:
                readme_content = f.read()
            with open(license_path) as f:
                license_content = f.read()
        except Exception as e:
            license_content = "License not found"
            readme_content = "README not found: Exception" + str(e)

        html_readme_content = markdown.markdown(
            readme_content,
            extensions=["fenced_code", "tables"],
        )
        html_license_content = markdown.markdown(license_content)
        context = {
            "readme_content": html_readme_content,
            "license_content": html_license_content,
        }
        return self._render_template("about.html", request, context)

    async def config(self, request: web.Request):
        button_actions: defaultdict[int, dict] = defaultdict(dict)
        for action in self._config.actions:
            button_actions[action.button_id]["action"] = action
        for longpress_action in self._config.longpress_action:
            button_actions[longpress_action.button_id]["longpress_action"] = (
                longpress_action
            )
        context = {
            "button_actions": button_actions,
            "leds": self._config.leds,
            "yaml_config": yaml.dump(self._config.model_dump(mode="json")),
        }
        return self._render_template("config.html", request, context)

    def set_version(self, version: str, hardware: str):
        self._version = version
        self._hardware = hardware

    async def device(self, request: web.Request):
        return self._render_template(
            "version.html",
            request,
            {"version": self._version, "hardware": self._hardware},
        )

    async def process(self):
        try:
            runner = web.AppRunner(self.app, access_log=None)
            await runner.setup()
            site = web.TCPSite(runner, self.host, self.port)
            await site.start()
            _logger.info("WebServer running at http://%s:%s", self.host, self.port)
            self._running = True
        except Exception as e:
            _logger.error("Error starting WebServer: %s", e)
            self._running = False

    async def stop(self):
        await self.app.shutdown()
        await self.app.cleanup()
        _logger.info("WebServer stopped")



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger <matthias@bilger.info>
import argparse  # Added import for argparse
import asyncio
import logging
import pathlib
import signal
import threading

import daiquiri
from mutenix.config import load_config
from mutenix.config import LoggingConfig
from mutenix.macropad import Macropad
from mutenix.tray_icon import run_trayicon
from mutenix.updates import check_for_self_update
from mutenix.utils import ensure_process_run_once
from mutenix.version import MAJOR
from mutenix.version import MINOR
from mutenix.version import PATCH

# Configure logging to write to a file
logging.basicConfig(level=logging.ERROR)
_logger = logging.getLogger(__name__)


def parse_arguments():
    parser = argparse.ArgumentParser(description="Mutenix Macropad Controller")
    parser.add_argument(
        "--config",
        type=pathlib.Path,
        help="Path to the configuration file",
    )
    parser.add_argument(
        "--update-file",
        type=str,
        help="Path to the update tar.gz file",
    )
    parser.add_argument(
        "--list-devices",
        action="store_true",
        help="List all connected devices",
    )
    return parser.parse_args()


def register_signal_handler(macropad: Macropad):
    """
    Registers a signal handler to shut down the Macropad gracefully on SIGINT.
    Args:
        macropad (Macropad): The Macropad instance to be shut down on SIGINT.
    """

    def signal_handler(signal, frame):  # pragma: no cover
        print("Shuting down...")
        _logger.info("SIGINT received, shutting down...")
        asyncio.run(macropad.stop())

    signal.signal(signal.SIGINT, signal_handler)


def list_devices():
    import hid

    for device in sorted(hid.enumerate(), key=lambda x: x["vendor_id"]):
        if "mutenix" in device["product_string"].lower():
            print("********** ")
        print(device)


def setup_logging(logging_config: LoggingConfig):
    log_file_path = logging_config.file_path or pathlib.Path.cwd() / "mutenix.log"
    log_level = logging_config.level.to_logging_level()
    all_levels = [log_level]
    outputs = []
    if logging_config.file_enabled:
        file_log_level = (
            logging_config.file_level.to_logging_level()
            if logging_config.file_level
            else log_level
        )
        outputs.append(
            daiquiri.output.RotatingFile(
                log_file_path,
                level=file_log_level,
                max_size_bytes=logging_config.file_max_size,
                backup_count=logging_config.file_backup_count,
            ),
        )
        all_levels.append(file_log_level)
    if logging_config.console_enabled:
        console_log_level = (
            logging_config.console_level.to_logging_level()
            if logging_config.console_level
            else log_level
        )
        outputs.append(
            daiquiri.output.Stream(
                level=console_log_level,
                formatter=daiquiri.formatter.ColorExtrasFormatter(
                    fmt="%(asctime)s - %(name)-25s [%(levelname)-8s]: %(message)s",
                ),
            ),
        )
        all_levels.append(console_log_level)
    daiquiri.setup(
        level=min(all_levels),
        outputs=outputs,
    )
    daiquiri.parse_and_set_default_log_levels(logging_config.submomdules)
    global _logger


@ensure_process_run_once()
def main(args: argparse.Namespace):
    config = load_config(args.config)
    print(config.logging)

    setup_logging(config.logging)

    if args.list_devices:
        return list_devices()

    check_for_self_update(MAJOR, MINOR, PATCH, config.proxy)
    macropad = Macropad(config)
    register_signal_handler(macropad)

    if args.update_file:
        _logger.info("Starting manual update with file: %s", args.update_file)
        asyncio.run(macropad.manual_update(args.update_file))
        return

    def run_asyncio_loop():  # pragma: no cover
        asyncio.run(macropad.process())

    _logger.info("Running Main Thread")
    loop_thread = threading.Thread(target=run_asyncio_loop)
    loop_thread.start()

    _logger.info("Tray icon start")
    run_trayicon(macropad)
    _logger.info("Tray icon stopped")

    loop_thread.join()
    _logger.info("Trhead joined")


def runmain():  # pragma: no cover
    args = parse_arguments()
    logging.basicConfig(level=logging.INFO)
    main(args)


if __name__ == "__main__":  # pragma: no cover
    runmain()



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger <matthias@bilger.info>
import json
import logging
import math
import os
import pathlib
import tarfile
import tempfile
import time
from collections.abc import Sequence
from typing import BinaryIO

import hid
import python_minifier
from mutenix.updates.chunks import Chunk
from mutenix.updates.chunks import Completed
from mutenix.updates.chunks import FileChunk
from mutenix.updates.chunks import FileDelete
from mutenix.updates.chunks import FileEnd
from mutenix.updates.chunks import FileStart
from mutenix.updates.constants import HID_COMMAND_PREPARE_UPDATE
from mutenix.updates.constants import HID_COMMAND_RESET
from mutenix.updates.constants import HID_REPORT_ID_COMMUNICATION
from mutenix.updates.constants import HID_REPORT_ID_TRANSFER
from mutenix.updates.constants import MAX_CHUNK_SIZE
from mutenix.updates.constants import STATE_CHANGE_SLEEP_TIME
from mutenix.updates.device_messages import ChunkAck
from mutenix.updates.device_messages import LogMessage
from mutenix.updates.device_messages import parse_hid_update_message
from mutenix.updates.device_messages import UpdateError
from tqdm import tqdm

_logger = logging.getLogger(__name__)


def perform_upgrade_with_file(device: hid.device, file_stream: BinaryIO):
    with tempfile.TemporaryDirectory() as tmpdirname:
        tmpdir = pathlib.Path(tmpdirname)
        with tarfile.open(fileobj=file_stream, mode="r:gz") as tar:
            tar.extractall(path=tmpdirname)
        files = list(
            map(
                lambda x: tmpdir / x,
                filter(
                    lambda x: (x.endswith(".py") or x.endswith(".delete"))
                    and not x.startswith("."),
                    os.listdir(tmpdirname),
                ),
            ),
        )
        _logger.debug("Updating device with files: %s", files)
        perform_hid_upgrade(device, files)
        _logger.info("Successfully updated device firmware")


class TransferFile:
    def __init__(self, file_id, filename: str | pathlib.Path):
        self.id = file_id
        file = pathlib.Path(filename) if isinstance(filename, str) else filename
        self.filename = file.name
        self.packages_sent: list[int] = []
        self._chunks: list[Chunk] = []
        if self.filename.endswith(".delete"):
            self.filename = self.filename[:-7]
            self._chunks = [FileDelete(self.id, self.filename)]
            return

        with open(file, "r") as f:
            if file.suffix == ".py":
                self.content = python_minifier.minify(
                    f.read(),
                    remove_annotations=True,
                    rename_globals=False,
                ).encode("utf-8")
            else:
                self.content = f.read().encode("utf-8")
        # Workaround for update issue
        self.size = len(self.content)
        self.content = self.content + b"\x20" * (
            MAX_CHUNK_SIZE - (len(self.content) % MAX_CHUNK_SIZE)
        )
        _logger.info("Size %s, %s", self.size, json.dumps(self.content.decode("utf-8")))
        self.make_chunks()
        _logger.debug("File %s has %s chunks", self.filename, len(self._chunks))

    def make_chunks(self):
        total_packages = self.calculate_total_packages()
        self.add_file_start_chunk(total_packages)
        self.add_file_chunks(total_packages)
        self.add_file_end_chunk()

    def calculate_total_packages(self):
        return math.ceil(self.size / MAX_CHUNK_SIZE)

    def add_file_start_chunk(self, total_packages):
        self._chunks.append(
            FileStart(self.id, 0, total_packages, self.filename, self.size),
        )

    def add_file_chunks(self, total_packages):
        for i in range(0, self.size, MAX_CHUNK_SIZE):
            self._chunks.append(
                FileChunk(
                    self.id,
                    i // MAX_CHUNK_SIZE,
                    total_packages,
                    self.content[i : i + MAX_CHUNK_SIZE],
                ),
            )

    def add_file_end_chunk(self):
        self._chunks.append(FileEnd(self.id))

    def get_next_chunk(self) -> Chunk | None:
        if self.is_complete():
            return None
        return next((chunk for chunk in self._chunks if not chunk.acked), None)

    def acknowledge_chunk(self, chunk: ChunkAck):
        # This line is excluded from coverage reports because it is a safeguard
        # and is unlikely to be met during normal execution.
        if chunk.id != self.id:  # pragma: no cover
            return
        acked_chunk = next(
            (
                x
                for x in self._chunks
                if x.type_ == chunk.type_ and x.package == chunk.package
            ),
            None,
        )
        if not acked_chunk:  # pragma: no cover
            _logger.warning("No chunk found for ack")
            return
        acked_chunk._acked = True
        _logger.debug("Acked chunk %s", chunk)

    @property
    def chunks(self):
        return len(self._chunks)

    def is_complete(self):
        return all(map(lambda x: x.acked, self._chunks))


def send_hid_command(device: hid.device, command: int):
    device.write([HID_REPORT_ID_COMMUNICATION, command] + [0] * 7)


def perform_hid_upgrade(device: hid.device, files: Sequence[str | pathlib.Path]):
    _logger.debug("Opening device for update")
    _logger.debug("Sending prepare update")
    send_hid_command(device, HID_COMMAND_PREPARE_UPDATE)
    time.sleep(STATE_CHANGE_SLEEP_TIME)

    transfer_files = [TransferFile(file_id, file) for file_id, file in enumerate(files)]

    _logger.debug("Preparing to send %s files", len(transfer_files))
    cancelled = False

    for i, file in enumerate(transfer_files, 1):
        if cancelled:
            break
        fileprogress = tqdm(
            total=file.chunks,
            desc=f"Sending file {file.filename:25} {i:2}/{len(transfer_files)}",
        )
        while True:
            received = device.read(100, 1000)
            if len(received) > 0:
                rcvd = parse_hid_update_message(bytes(received[1:]))

                if isinstance(rcvd, ChunkAck):
                    ack_file = next(
                        (f for f in transfer_files if f.id == rcvd.id),
                        None,
                    )
                    if not ack_file:
                        _logger.warning("No file id found for ack")
                        continue
                    fileprogress.update(1)
                    ack_file.acknowledge_chunk(rcvd)
                elif isinstance(rcvd, UpdateError):
                    print("Error received from device: ", rcvd)
                    _logger.error("Error received from device: %s", rcvd)
                    cancelled = True
                    break
                elif isinstance(rcvd, LogMessage):
                    print(rcvd)

            chunk = file.get_next_chunk()
            if not chunk:
                fileprogress.close()
                break
            _logger.debug(
                "Sending chunk (%s...) of file %s",
                chunk.packet()[:10],
                file.filename,
            )
            cnk = bytes((HID_REPORT_ID_TRANSFER,)) + chunk.packet()
            try:
                device.write(cnk)
            except Exception as e:
                _logger.error("Failed to write chunk to device: %s", e)
                cancelled = True
                break

    time.sleep(STATE_CHANGE_SLEEP_TIME)
    try:
        device.write(bytes((HID_REPORT_ID_TRANSFER,)) + Completed().packet())
    except Exception as e:
        _logger.error("Failed to write Completed packet to device: %s", e)
    time.sleep(STATE_CHANGE_SLEEP_TIME)
    print("Resetting")
    _logger.info("Resetting")
    send_hid_command(device, HID_COMMAND_RESET)



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger <matthias@bilger.info>
import io
import logging
import webbrowser

import hid
import requests
import semver
from mutenix.hid_commands import VersionInfo
from mutenix.updates.device_update import perform_upgrade_with_file

_logger = logging.getLogger(__name__)


def check_for_device_update(
    device: hid.device,
    device_version: VersionInfo,
    proxy: str | None = None,
):
    if proxy:
        proxies = {"https": proxy}
    else:
        proxies = {}
    try:
        result = requests.get(
            "https://api.github.com/repos/mutenix-org/firmware-macroboard/releases/latest",
            timeout=4,
            proxies=proxies,
        )
        if result.status_code != 200:
            _logger.error(
                "Failed to fetch latest release info, status code: %s",
                result.status_code,
            )
            return False

        releases = result.json()
        latest_version = releases.get("tag_name", "v0.0.0")[1:]
        _logger.debug("Latest version: %s", latest_version)
        online_version = semver.Version.parse(latest_version)
        local_version = semver.Version.parse(device_version.version)
        if online_version.compare(local_version) <= 0:
            _logger.info("Device is up to date")
            return False

        print("Device update available, starting update, please be patient")
        assets = releases.get("assets", [])
        for asset in assets:
            if asset.get("name") == f"v{latest_version}.tar.gz":
                update_url = asset.get("browser_download_url")
                result = requests.get(update_url)
                result.raise_for_status()
                perform_upgrade_with_file(device, io.BytesIO(result.content))
                return True
    except requests.RequestException as e:
        _logger.error("Failed to check for device update availability %s", e)


# region: Update Application
def check_for_self_update(major: int, minor: int, patch: int, proxy: str | None = None):
    if proxy:
        proxies = {"https": proxy}
    else:
        proxies = {}
    try:
        result = requests.get(
            "https://api.github.com/repos/mutenix-org/software-host/releases/latest",
            timeout=4,
            proxies=proxies,
        )
        if result.status_code != 200:
            _logger.error(
                "Failed to fetch latest release info, status code: %s",
                result.status_code,
            )
            return

        releases = result.json()
        latest_version = releases.get("tag_name", "v0.0.0")[1:]
        _logger.debug("Latest version: %s", latest_version)
        online_version = semver.Version.parse(latest_version)
        local_version = semver.Version(major=major, minor=minor, patch=patch)
        if online_version.compare(local_version) <= 0:
            _logger.info("Host Software is up to date")
            return

        _logger.info("Application update available, but auto update is disabled")
        webbrowser.open(releases.get("html_url"))
    except requests.RequestException as e:
        _logger.error("Failed to check for application update availability: %s", e)


# endregion



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger matthias@bilger.info
from __future__ import annotations

import asyncio
from unittest.mock import ANY
from unittest.mock import AsyncMock
from unittest.mock import patch

import pytest
from aiohttp.test_utils import AioHTTPTestCase
from mutenix.hid_commands import HidOutputMessage
from mutenix.hid_commands import LedColor
from mutenix.hid_commands import SetLed
from mutenix.hid_commands import Status
from mutenix.virtual_macropad import UnsupportedMessageTypeError
from mutenix.virtual_macropad import VirtualMacropad


class TestVirtualMacropad(AioHTTPTestCase):
    async def get_application(self):
        self.macropad = VirtualMacropad()
        return self.macropad.app

    def start_process(self):
        async def start_process():
            while True:
                await self.macropad.process()
                await asyncio.sleep(0.1)

        self.loop.create_task(start_process())

    async def test_index(self):
        request = await self.client.request("GET", "/")
        assert request.status == 200
        text = await request.text()
        assert "<!DOCTYPE html PUBLIC" in text

    async def test_button_handler(self):
        data = {"button": 1}
        request = await self.client.request("POST", "/button", json=data)
        assert request.status == 200

    async def test_send_msg(self):
        msg = SetLed(id=1, led_color=LedColor.RED)
        await self.macropad.send_msg(msg)
        async with self.macropad._led_status_lock:
            assert self.macropad._led_status[1] == "red"

    async def test_send_msg_invalid(self):
        class SomeMessage(HidOutputMessage):
            pass

        msg = SomeMessage()
        with pytest.raises(UnsupportedMessageTypeError):
            await self.macropad.send_msg(msg)

    async def test_process(self):
        await self.macropad.process()
        assert self.macropad.host == "127.0.0.1"
        assert self.macropad.port == 12909

    async def test_websocket_handler_button_press(self):
        self.start_process()
        ws = await self.client.ws_connect("/ws")
        await ws.send_json({"command": "button", "button": 1})
        await ws.close()

    async def test_websocket_handler_state_request(self):
        self.start_process()
        self.macropad._led_status[1] = "red"
        ws = await self.client.ws_connect("/ws")
        await ws.send_json({"command": "state_request"})
        msg = await ws.receive_json()
        assert "button" in msg
        assert "color" in msg
        await ws.close()

    async def test_websocket_handler_unknown(self):
        self.start_process()
        self.macropad._led_status[1] = "red"
        ws = await self.client.ws_connect("/ws")
        await ws.send_json({"command": "something_else"})
        msg = await ws.receive_json()
        assert "error" in msg
        await ws.close()

    async def test_websocket_handler_multiple_clients(self):
        self.start_process()
        self.macropad._led_status[1] = "red"
        ws1 = await self.client.ws_connect("/ws")
        ws2 = await self.client.ws_connect("/ws")
        await self.macropad.send_msg(SetLed(id=2, led_color=LedColor.GREEN))
        msg1 = await ws1.receive_json()
        msg2 = await ws2.receive_json()
        assert msg1["button"] == 2
        assert msg2["button"] == 2
        await ws1.close()
        await ws2.close()

    @pytest.mark.asyncio
    async def test_send_json_safe_success(self):
        ws = AsyncMock()
        data = {"button": 1, "color": "red"}
        await self.macropad._send_json_safe(ws, data)
        ws.send_json.assert_called_once_with(data)

    @pytest.mark.asyncio
    async def test_send_json_safe_failure(self):
        ws = AsyncMock()
        ws.send_json.side_effect = Exception("Test exception")
        data = {"button": 1, "color": "red"}
        with patch("mutenix.virtual_macropad._logger.error") as mock_logger_error:
            await self.macropad._send_json_safe(ws, data)
            ws.send_json.assert_called_once_with(data)
            mock_logger_error.assert_called_once_with(
                "Error sending LED status: %s to websocket %s",
                ANY,
                ANY,
            )

    async def test_register_callback(self):
        callback = AsyncMock()
        self.macropad.register_callback(callback)
        await self.macropad._handle_msg(Status.trigger_button(1))
        callback.assert_called_once()

    @pytest.mark.asyncio
    async def test_stop(self):
        self.macropad.app.shutdown = AsyncMock()
        self.macropad.app.cleanup = AsyncMock()
        await self.macropad.stop()
        self.macropad.app.shutdown.assert_called_once()
        self.macropad.app.cleanup.assert_called_once()

    async def test_favicon_32(self):
        request = await self.client.request("GET", "/favicon/32")
        assert request.status == 200
        assert request.content_type == "image/png"

    async def test_favicon_16(self):
        request = await self.client.request("GET", "/favicon/16")
        assert request.status == 200
        assert request.content_type == "image/png"

    async def test_favicon_apple_touch(self):
        request = await self.client.request("GET", "/favicon/apple_touch")
        assert request.status == 200
        assert request.content_type == "image/png"

    async def test_favicon_not_found(self):
        request = await self.client.request("GET", "/favicon/non_existent")
        assert request.status == 404

    async def test_favicon_svg(self):
        request = await self.client.request("GET", "/favicon.svg")
        assert request.status == 200
        assert request.content_type == "image/svg+xml"

    async def test_serve_manifest(self):
        request = await self.client.request("GET", "/site.webmanifest")
        assert request.status == 200
        manifest = await request.json()
        assert manifest["name"] == "Mutenix Virtual Macropad"
        assert manifest["short_name"] == "Mutenix"
        assert manifest["start_url"] == "/"
        assert manifest["display"] == "standalone"
        assert len(manifest["icons"]) == 4

    async def test_help(self):
        request = await self.client.request("GET", "/help")
        assert request.status == 200
        text = await request.text()
        assert "Help" in text

    async def test_about(self):
        request = await self.client.request("GET", "/about")
        assert request.status == 200
        text = await request.text()
        assert "About" in text

    async def test_websocket_handler_non_text_message(self):
        self.start_process()
        ws = await self.client.ws_connect("/ws")
        await ws.send_bytes(b"binary data")
        msg = await ws.receive_json()
        assert "error" in msg
        assert msg["error"] == "unknown message"
        await ws.close()

    async def test_led_handler(self):
        data = {"button": 1, "color": "blue"}
        request = await self.client.request("POST", "/led", json=data)
        assert request.status == 200
        assert self.macropad._led_input_status[1] == "blue"

    async def test_led_handler_invalid_data(self):
        data = {"button": "invalid", "color": "blue"}
        request = await self.client.request("POST", "/led", json=data)
        assert (
            request.status == 500
        )  # Assuming the handler does not handle invalid data gracefully

    async def test_get_led_status(self):
        self.macropad._led_input_status[1] = "blue"
        assert self.macropad.get_led_status(1) == "blue"
        assert self.macropad.get_led_status(2) == "black"



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger <matthias@bilger.info>
import os

from mutenix.utils.linux import ensure_process_run_once


def bring_teams_to_foreground() -> None:  # pragma: no cover
    """
    Bring the Microsoft Teams window to the foreground.

    This function attempts to bring the Microsoft Teams application window to the foreground
    on different operating systems (Windows, macOS, and Linux). It uses platform-specific
    methods to achieve this.

    On macOS, it uses AppleScript commands to activate the application and set it as frontmost.
    """
    os.system("osascript -e 'tell application \"Microsoft Teams\" to activate'")
    os.system(
        'osascript -e \'tell application "System Events" to tell process "Microsoft Teams" to set frontmost to true\'',
    )


__all__ = ["bring_teams_to_foreground", "ensure_process_run_once"]



import logging

_logger = logging.getLogger(__name__)


class UpdateError:
    def __init__(self, data: bytes):
        self.info = ""
        self.parse(data)

    def parse(self, data: bytes):
        self.identifier = data[:2].decode("utf-8")
        if not self.is_valid:
            return
        length = max(int.from_bytes(data[2:3], "little"), 33)
        self.info = data[3 : 3 + length].decode("utf-8")
        _logger.info("Error received: %s", self)

    @property
    def is_valid(self):
        return self.identifier == "ER"

    def __str__(self):
        if self.is_valid:
            return f"Error: {self.info}"
        return "Invalid Request"


class ChunkAck:
    def __init__(self, data: bytes):
        self.id = 0
        self.package = 0
        self.type_ = 0
        self.parse(data)

    def parse(self, data: bytes):
        self.identifier = data[:2].decode("utf-8")
        if not self.is_valid:
            return
        self.id = int.from_bytes(data[2:4], "little")
        self.package = int.from_bytes(data[4:6], "little")
        self.type_ = int.from_bytes(data[6:7], "little")

    @property
    def is_valid(self):
        return self.identifier == "AK"

    def __str__(self):
        if self.is_valid:
            return f"File: {self.id}, Type: {self.type_}, Package: {self.package}"
        return "Invalid Request"


class LogMessage:
    def __init__(self, data: bytes):
        self.message = ""
        self.parse(data)

    def parse(self, data: bytes):
        self.identifier = data[:2].decode("utf-8")
        if not self.is_valid:
            return
        self.level = "debug" if self.identifier == "LD" else "error"
        self.message = data[2 : data.index(0)].decode("utf-8")

    @property
    def is_valid(self):
        return self.identifier in ("LE", "LD")

    def __str__(self):
        if self.is_valid:
            return f"{self.level}: {self.message}"
        return "Invalid Request"


def parse_hid_update_message(data: bytes) -> ChunkAck | UpdateError | LogMessage | None:
    if len(data) < 2:
        return None
    val = data[:2].decode("utf-8")
    match val:
        case "AK":
            return ChunkAck(data)
        case "ER":
            return UpdateError(data)
        case "LD", "LE":
            return LogMessage(data)
    return None



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger <matthias@bilger.info>
import asyncio
import json
import logging
from typing import Callable

from aiohttp import web
from mutenix.hid_commands import HidOutputMessage
from mutenix.hid_commands import SetLed
from mutenix.hid_commands import Status
from mutenix.web_server import WebServer

HOST = "127.0.0.1"
PORT = 12909

_logger = logging.getLogger(__name__)


class UnsupportedMessageTypeError(Exception):
    """Exception raised for unsupported message types in VirtualMacropad."""

    pass


class VirtualMacropad(WebServer):
    """A virtual representation of the Macropad for testing or playing around."""

    def __init__(self, host: str = HOST, port: int = PORT):
        super().__init__(host, port)
        self._callbacks: list[Callable[[HidOutputMessage], asyncio.Future]] = []
        self._websockets: set[web.WebSocketResponse] = set()
        self._led_status: dict[int, str] = {}
        self._led_input_status: dict[int, str] = {}
        self._led_status_lock = asyncio.Lock()

        self.app.add_routes(
            [
                web.post("/button", self.button_handler),
                web.get("/ws", self.websocket_handler),
                web.post("/led", self.led_handler),
            ],
        )

    def register_callback(self, callback: Callable[[HidOutputMessage], asyncio.Future]):
        self._callbacks.append(callback)

    async def button_handler(self, request: web.Request):
        data = await request.json()
        await self._handle_msg(Status.trigger_button(data.get("button")))
        return web.Response(status=200)

    async def _handle_msg(self, msg: HidOutputMessage):
        for callback in self._callbacks:
            await callback(msg)

    async def handle_state_request(self, ws):
        async with self._led_status_lock:
            for i, color in self._led_status.items():
                if color:
                    await ws.send_json({"button": i, "color": color})

    async def websocket_handler(self, request: web.Request):
        ws = web.WebSocketResponse()
        await ws.prepare(request)
        self._websockets.add(ws)
        async for msg in ws:
            if msg.type == web.WSMsgType.TEXT:
                data = json.loads(msg.data)
                match data["command"]:
                    case "button":
                        await self._handle_msg(
                            Status.trigger_button(data.get("button")),
                        )
                    case "state_request":
                        await self.handle_state_request(ws)
                    case _:
                        _logger.info("Unknown message: %s", data)
                        await ws.send_json({"error": "unknown command"})
            else:
                _logger.info("Unknown message: %s", msg)
                await ws.send_json({"error": "unknown message"})
        self._websockets.remove(ws)
        return ws

    @staticmethod
    async def _send_json_safe(ws, data):
        try:
            await ws.send_json(data)
        except Exception as e:
            _logger.error("Error sending LED status: %s to websocket %s", e, ws)

    def _send_led_status(self, button: int, color: str):
        for ws in self._websockets:
            asyncio.create_task(
                self._send_json_safe(ws, {"button": button, "color": color}),
            )

    async def send_msg(self, msg: HidOutputMessage):
        if isinstance(msg, SetLed):
            color = msg.color.name.lower()
            async with self._led_status_lock:
                self._led_status[msg.id] = color
            self._send_led_status(msg.id, color)
        else:
            raise UnsupportedMessageTypeError("Unsupported message type")
        _logger.debug("Sent message: %s", msg)

    async def led_handler(self, request: web.Request):
        data = await request.json()
        self._led_input_status[int(data["button"])] = data["color"]
        return web.Response(status=200)

    def get_led_status(self, button: int) -> str:
        return self._led_input_status.get(button, "black")



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger <matthias@bilger.info>
import logging
import os
import pathlib
import subprocess
import sys
import tempfile

import psutil

_logger = logging.getLogger(__name__)


def bring_teams_to_foreground() -> None:  # pragma: no cover
    try:
        # Get the window ID of Microsoft Teams
        window_id = (
            subprocess.check_output(
                "xdotool search --name 'Microsoft Teams'",
                shell=True,
            )
            .strip()
            .decode()
        )
        # Activate the window
        os.system(f"xdotool windowactivate {window_id}")
    except Exception as e:
        _logger.error("Microsoft Teams window not found: %s", e)


def ensure_process_run_once(
    lockfile_path: pathlib.Path = pathlib.Path(tempfile.gettempdir()),
):
    def outerwrapper(func):
        def wrapper(*args, **kwargs):
            lock_file = lockfile_path / "mutenix.lock"
            _logger.info("Using Lock file: %s", lock_file)
            if lock_file.exists():
                _logger.debug("Lock file exists. Another instance might be running.")
                try:
                    with lock_file.open("r") as f:
                        pid = int(f.read().strip())
                    _logger.info("Found PID %s in lock file", pid)
                    if psutil.pid_exists(pid):
                        _logger.error(
                            "The other instance %s is still running, exiting this one",
                            pid,
                        )
                        sys.exit(1)
                except (OSError, ValueError):
                    _logger.info("Stale lock file found. Removing and continuing.")
                _logger.info("Removing lockfile")
                lock_file.unlink()
            with lock_file.open("w") as f:
                f.write(str(os.getpid()))
            try:
                result = func(*args, **kwargs)
            finally:
                lock_file.unlink()
            return result

        return wrapper

    return outerwrapper



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger matthias@bilger.info
from __future__ import annotations

from pathlib import Path
from unittest.mock import mock_open
from unittest.mock import patch

from mutenix.tray_icon import load_image


def test_load_image_success():
    file_name = "test_image.png"
    file_path = Path(__file__).parent.parent / "src" / "mutenix" / "assets" / file_name

    with patch(
        "mutenix.tray_icon.Path.open",
        mock_open(read_data="image_data"),
    ):
        with patch("PIL.Image.open") as mock_image_open:
            load_image(file_name)
            mock_image_open.assert_called_once_with(file_path)


def test_load_image_file_not_found():
    file_name = "non_existent_image.png"
    load_image(file_name)



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger matthias@bilger.info
from __future__ import annotations

import os
import pathlib
import unittest
from unittest.mock import MagicMock
from unittest.mock import mock_open
from unittest.mock import patch

import requests
from mutenix.hid_commands import HardwareTypes
from mutenix.updates import check_for_device_update
from mutenix.updates import check_for_self_update
from mutenix.updates import VersionInfo
from mutenix.updates.chunks import Chunk
from mutenix.updates.chunks import FileChunk
from mutenix.updates.chunks import FileDelete
from mutenix.updates.chunks import FileEnd
from mutenix.updates.chunks import FileStart
from mutenix.updates.constants import MAX_CHUNK_SIZE
from mutenix.updates.device_messages import ChunkAck
from mutenix.updates.device_messages import UpdateError
from mutenix.updates.device_update import perform_hid_upgrade
from mutenix.updates.device_update import TransferFile


class TestUpdates(unittest.TestCase):
    @patch("mutenix.updates.requests.get")
    @patch("mutenix.updates.semver.compare")
    def test_check_for_device_update_up_to_date(self, mock_compare, mock_get):
        mock_compare.return_value = 0
        mock_get.return_value.status_code = 200
        mock_get.return_value.json.return_value = {"tag_name": "v1.0.0"}
        device_version = VersionInfo(
            buffer=bytes([1, 0, 0, HardwareTypes.UNKNOWN.value, 0, 0, 0, 0]),
        )
        mock_device = MagicMock()
        check_for_device_update(mock_device, device_version)

        mock_get.assert_called_once()

    @patch("mutenix.updates.requests.get")
    @patch("mutenix.updates.semver.compare")
    @patch("mutenix.updates.device_update.perform_hid_upgrade")
    def test_check_for_device_update_needs_update(
        self,
        mock_upgrade,
        mock_compare,
        mock_get,
    ):
        mock_compare.return_value = -1
        mock_get.return_value.status_code = 200
        mock_get.return_value.json.return_value = {
            "tag_name": "v2.0.0",
            "assets": [
                {
                    "name": "v2.0.0.tar.gz",
                    "browser_download_url": "http://example.com/update.tar.gz",
                },
            ],
        }

        mock_update_response = MagicMock()
        mock_update_response.status_code = 200
        mock_update_response.content = b"fake content"
        mock_get.side_effect = [mock_get.return_value, mock_update_response]

        device_version = VersionInfo(
            buffer=bytes([1, 0, 0, HardwareTypes.UNKNOWN.value, 0, 0, 0, 0]),
        )
        with patch("tarfile.open") as mock_tarfile:
            mock_tarfile.return_value.__enter__.return_value.extractall = MagicMock()
            mock_device = MagicMock()
            check_for_device_update(mock_device, device_version)

        mock_get.assert_called()
        mock_upgrade.assert_called_once()

    @patch("mutenix.updates.requests.get")
    def test_check_for_device_update_no_response(self, mock_get):
        mock_get.return_value.status_code = 500
        mock_get.return_value.json.return_value = None
        device_version = VersionInfo(
            buffer=bytes([1, 0, 0, HardwareTypes.UNKNOWN.value, 0, 0, 0, 0]),
        )
        mock_device = MagicMock()
        check_for_device_update(mock_device, device_version)

        mock_get.assert_called_once()

    @patch("mutenix.updates.requests.get")
    @patch("mutenix.updates.semver.compare")
    @patch("mutenix.updates.device_update.perform_hid_upgrade")
    def test_check_for_device_update_needs_update_but_fails(
        self,
        mock_upgrade,
        mock_compare,
        mock_get,
    ):
        mock_compare.return_value = -1
        mock_get.return_value.status_code = 200
        mock_get.return_value.json.return_value = {
            "tag_name": "v2.0.0",
            "assets": [
                {
                    "name": "v2.0.0",
                    "browser_download_url": "http://example.com/update.tar.gz",
                },
            ],
        }

        mock_update_response = MagicMock()
        mock_update_response.side_effect = requests.RequestException("Network error")
        mock_get.side_effect = [
            mock_get.return_value,
            requests.RequestException("Network error"),
        ]

        device_version = VersionInfo(
            buffer=bytes([1, 0, 0, HardwareTypes.UNKNOWN.value, 0, 0, 0, 0]),
        )

        mock_device = MagicMock()
        check_for_device_update(mock_device, device_version)

        mock_get.assert_called()
        mock_upgrade.assert_not_called()

    @patch("mutenix.updates.requests.get")
    @patch("mutenix.updates.semver.compare")
    def test_check_for_self_update_up_to_date(self, mock_compare, mock_get):
        mock_compare.return_value = 0
        mock_get.return_value.status_code = 200
        mock_get.return_value.json.return_value = {"tag_name": "v1.0.0"}

        check_for_self_update(1, 0, 0)

        mock_get.assert_called_once()

    @patch("mutenix.updates.requests.get")
    def test_check_for_device_update_request_exception(self, mock_get):
        mock_get.side_effect = requests.RequestException("Network error")
        device_version = VersionInfo(
            buffer=bytes([1, 0, 0, HardwareTypes.UNKNOWN.value, 0, 0, 0, 0]),
        )
        mock_device = MagicMock()

        with self.assertLogs("mutenix.updates", level="ERROR") as log:
            check_for_device_update(mock_device, device_version)

        mock_get.assert_called_once()
        self.assertIn("Failed to check for device update availability", log.output[0])

    @patch("mutenix.updates.requests.get")
    @patch("mutenix.updates.semver.compare")
    def test_check_for_self_update_needs_update(self, mock_compare, mock_get):
        mock_compare.return_value = -1
        mock_get.return_value.status_code = 200
        mock_get.return_value.json.return_value = {
            "tag_name": "v2.0.0",
            "html_url": "http://example.com",
        }

        mock_update_response = MagicMock()
        mock_update_response.status_code = 200
        mock_update_response.content = b"fake content"
        mock_get.side_effect = [mock_get.return_value, mock_update_response]

        with patch("webbrowser.open"):
            check_for_self_update(1, 0, 0)

        mock_get.assert_called()

    @patch("mutenix.updates.device_update.hid.device")
    @patch("python_minifier.minify", side_effect=lambda x, *args, **kwargs: str(x))
    def test_perform_hid_upgrade_success(self, mock_device, mock_minify):
        mock_device_instance = MagicMock()
        mock_device.return_value = mock_device_instance

        mock_device_instance.read.side_effect = [
            bytes(),
            bytes([2, 65, 75, 0, 0, 0, 0, 1, 0, 0]),  # Acknowledge for first file
            bytes([2, 65, 75, 0, 0, 0, 0, 2, 0, 0]),  # Acknowledge for first file
            bytes([2, 65, 75, 0, 0, 0, 0, 3, 0, 0]),  # Acknowledge for first file
            bytes([2, 65, 75, 1, 0, 0, 0, 1, 0, 0]),  # Acknowledge for first file
            bytes([2, 65, 75, 1, 0, 0, 0, 2, 0, 0]),  # Acknowledge for first file
            bytes([2, 65, 75, 1, 0, 0, 0, 3, 0, 0]),  # Acknowledge for first file
            bytes([2, 65, 75, 2, 0, 0, 0, 1, 0, 0]),  # Acknowledge for first file
            bytes([2, 65, 75, 2, 0, 0, 0, 2, 0, 0]),  # Acknowledge for first file
            bytes([2, 65, 75, 2, 0, 0, 0, 3, 0, 0]),  # Acknowledge for first file
            bytes(),
            bytes(),
            bytes(),
            bytes(),
            bytes(),
            bytes(),
        ]

        with patch("mutenix.updates.constants.DATA_TRANSFER_SLEEP_TIME", 0.0001):
            with patch("mutenix.updates.constants.STATE_CHANGE_SLEEP_TIME", 0.0001):
                with patch("builtins.open", mock_open(read_data=b"fake content")):
                    with patch("pathlib.Path.is_file", return_value=True):
                        with patch(
                            "pathlib.Path.open",
                            mock_open(read_data=b"fake content"),
                        ):
                            perform_hid_upgrade(
                                mock_device_instance,
                                ["file1.py", "file2.py", "file3.py"],
                            )

        self.assertEqual(
            mock_device_instance.write.call_count,
            10,
        )  # 3 files * 3 chunks each + 1 state change commands

    @patch("mutenix.updates.device_update.hid.device")
    @patch("python_minifier.minify", side_effect=lambda x, *args, **kwargs: str(x))
    def test_perform_hid_upgrade_ack(self, mock_device, mock_minify):
        mock_device_instance = MagicMock()
        mock_device.return_value = mock_device_instance

        mock_device_instance.read.side_effect = [
            bytes(),  # No more requests
            bytes([2, 65, 75, 0, 0, 0, 0, 1, 0, 0]),  # Acknowledge for first file
            bytes([2, 65, 75, 0, 0, 0, 0, 2, 0, 0]),  # Acknowledge for first file
            bytes([2, 65, 75, 0, 0, 0, 0, 3, 0, 0]),  # Acknowledge for first file
            bytes(),  # No more requests
            bytes(),  # No more requests
            bytes(),  # No more requests
            bytes(),  # No more requests
            bytes(),  # No more requests
            bytes(),  # No more requests
        ]

        with patch("mutenix.updates.constants.DATA_TRANSFER_SLEEP_TIME", 0.0001):
            with patch("mutenix.updates.constants.STATE_CHANGE_SLEEP_TIME", 0.0001):
                with patch("builtins.open", mock_open(read_data=b"fake content")):
                    with patch("pathlib.Path.is_file", return_value=True):
                        with patch(
                            "pathlib.Path.open",
                            mock_open(read_data=b"fake content"),
                        ):
                            perform_hid_upgrade(mock_device_instance, ["file1.py"])

    @patch("mutenix.updates.requests.get")
    def test_check_for_self_update_request_error(self, mock_get):
        mock_get.side_effect = requests.RequestException("Network error")

        with self.assertLogs("mutenix.updates", level="ERROR") as log:
            check_for_self_update(1, 0, 0)

        self.assertIn(
            "Failed to check for application update availability",
            log.output[0],
        )

    @patch("mutenix.updates.requests.get")
    def test_check_for_self_update_status_code_error(self, mock_get):
        mock_get.return_value.status_code = 500

        with self.assertLogs("mutenix.updates", level="ERROR") as log:
            check_for_self_update(1, 0, 0)

        self.assertIn(
            "Failed to fetch latest release info, status code: 500",
            log.output[0],
        )


class TestFileChunk(unittest.TestCase):
    def test_file_chunk_packet(self):
        chunk = FileChunk(1, 2, 3, b"content")
        packet = chunk.packet()
        self.assertEqual(packet[:2], (2).to_bytes(2, "little"))
        self.assertEqual(packet[2:4], (1).to_bytes(2, "little"))
        self.assertEqual(packet[4:6], (3).to_bytes(2, "little"))
        self.assertEqual(packet[6:8], (2).to_bytes(2, "little"))
        self.assertEqual(packet[8:16], b"content" + b"\0")


class TestFileStart(unittest.TestCase):
    def test_file_start_packet(self):
        start = FileStart(1, 0, 3, "test.py", 100)
        packet = start.packet()
        self.assertEqual(packet[:2], (1).to_bytes(2, "little"))
        self.assertEqual(packet[2:4], (1).to_bytes(2, "little"))
        self.assertEqual(packet[4:6], (3).to_bytes(2, "little"))
        self.assertEqual(packet[6:8], (0).to_bytes(2, "little"))
        self.assertEqual(packet[8:9], bytes((7,)))
        self.assertEqual(
            packet[9:19],
            b"test.py" + bytes((2,)) + (100).to_bytes(2, "little"),
        )


class TestFileEnd(unittest.TestCase):
    def test_file_end_packet(self):
        end = FileEnd(1)
        packet = end.packet()
        self.assertEqual(packet[:2], (3).to_bytes(2, "little"))
        self.assertEqual(packet[2:4], (1).to_bytes(2, "little"))
        self.assertEqual(packet[4:], b"\0" * (MAX_CHUNK_SIZE + 4))


class TestTransferFile(unittest.TestCase):
    def setUp(self):
        self.file_content = b"fake content" * 10
        self.file_path = "test_file.py"
        with open(self.file_path, "wb") as f:
            f.write(self.file_content)

    def tearDown(self):
        os.remove(self.file_path)

    @patch("python_minifier.minify", side_effect=lambda x, *args, **kwargs: str(x))
    def test_transfer_file_chunks(self, mock_minify):
        transfer_file = TransferFile(1, self.file_path)
        self.assertGreaterEqual(transfer_file.size, len(self.file_content))
        self.assertEqual(len(transfer_file._chunks), transfer_file.chunks)

    @patch("python_minifier.minify", side_effect=lambda x, *args, **kwargs: str(x))
    def test_transfer_file_get_next_chunk(self, mock_minify):
        transfer_file = TransferFile(1, self.file_path)
        chunk = transfer_file.get_next_chunk()
        self.assertIsInstance(chunk, Chunk)

    @patch("python_minifier.minify", side_effect=lambda x, *args, **kwargs: str(x))
    def test_transfer_file_is_complete(self, mock_minify):
        transfer_file = TransferFile(1, self.file_path)
        while not transfer_file.is_complete():
            chunk = transfer_file.get_next_chunk()
            if chunk:
                dfc = ChunkAck(
                    b"AK"
                    + (chunk.id).to_bytes(2, "little")
                    + (chunk.package).to_bytes(2, "little")
                    + (chunk.type_).to_bytes(1, "little")
                    + b"\0" * 2,
                )
                dfc.type_ = chunk.type_
                dfc.package = chunk.package
                transfer_file.acknowledge_chunk(dfc)
        self.assertTrue(transfer_file.is_complete())

    @patch("python_minifier.minify", side_effect=lambda x, *args, **kwargs: str(x))
    def test_transfer_file_from_path(self, mock_minify):
        transfer_file = TransferFile(1, pathlib.Path(self.file_path))
        self.assertEqual(transfer_file.filename, "test_file.py")
        self.assertGreaterEqual(transfer_file.size, len(self.file_content))


class TestPerformHidUpgradeError(unittest.TestCase):
    @patch("mutenix.updates.hid.device")
    @patch("python_minifier.minify", side_effect=lambda x, *args, **kwargs: str(x))
    def test_perform_hid_upgrade_error(self, mock_device, mock_minify):
        mock_device_instance = MagicMock()
        mock_device.return_value = mock_device_instance

        mock_device_instance.read.side_effect = [
            bytes([2, 69, 82, 5, 69, 114, 114, 111, 114]),  # Error
        ]

        with patch("mutenix.updates.constants.DATA_TRANSFER_SLEEP_TIME", 0.0001):
            with patch("mutenix.updates.constants.STATE_CHANGE_SLEEP_TIME", 0.0001):
                with patch("builtins.open", mock_open(read_data=b"fake content")):
                    with patch("pathlib.Path.is_file", return_value=True):
                        with patch(
                            "pathlib.Path.open",
                            mock_open(read_data=b"fake content"),
                        ):
                            with self.assertLogs(
                                "mutenix.updates",
                                level="ERROR",
                            ) as log:
                                perform_hid_upgrade(mock_device_instance, ["file1.py"])

        self.assertIn("Error received from device: Error", log.output[0])


class TestTransferFileInit(unittest.TestCase):
    def setUp(self):
        self.file_content = b"fake content" * 10
        self.file_path = "test_file.py"
        with open(self.file_path, "wb") as f:
            f.write(self.file_content)

    def tearDown(self):
        os.remove(self.file_path)

    @patch("python_minifier.minify", side_effect=lambda x, *args, **kwargs: str(x))
    def test_transfer_file_init(self, mock_minify):
        transfer_file = TransferFile(1, self.file_path)
        self.assertEqual(transfer_file.id, 1)
        self.assertEqual(transfer_file.filename, "test_file.py")
        self.assertGreaterEqual(transfer_file.size, len(self.file_content))
        self.assertEqual(len(transfer_file._chunks), transfer_file.chunks)

    def test_transfer_file_init_delete(self):
        delete_file_path = "test_file.py.delete"
        with open(delete_file_path, "wb") as f:
            f.write(self.file_content)
        transfer_file = TransferFile(1, delete_file_path)
        self.assertEqual(transfer_file.id, 1)
        self.assertEqual(transfer_file.filename, "test_file.py")
        self.assertEqual(len(transfer_file._chunks), 1)
        self.assertIsInstance(transfer_file._chunks[0], FileDelete)
        os.remove(delete_file_path)


class TestChunkAck(unittest.TestCase):
    def test_chunk_ack_str_valid(self):
        data = (
            b"AK"
            + (1).to_bytes(2, "little")
            + (2).to_bytes(2, "little")
            + (3).to_bytes(1, "little")
        )
        chunk_ack = ChunkAck(data)
        self.assertTrue(chunk_ack.is_valid)
        self.assertEqual(str(chunk_ack), "File: 1, Type: 3, Package: 2")

    def test_chunk_ack_str_invalid(self):
        data = (
            b"XX"
            + (1).to_bytes(2, "little")
            + (2).to_bytes(2, "little")
            + (3).to_bytes(1, "little")
        )
        chunk_ack = ChunkAck(data)
        self.assertFalse(chunk_ack.is_valid)
        self.assertEqual(str(chunk_ack), "Invalid Request")
        self.assertEqual(str(chunk_ack), "Invalid Request")


class TestUpdateError(unittest.TestCase):
    def test_update_error_str_valid(self):
        data = b"ER" + (5).to_bytes(1, "little") + b"Error"
        update_error = UpdateError(data)
        self.assertTrue(update_error.is_valid)
        self.assertEqual(str(update_error), "Error: Error")

    def test_update_error_str_invalid(self):
        data = b"XX" + (5).to_bytes(1, "little") + b"Error"
        update_error = UpdateError(data)
        self.assertFalse(update_error.is_valid)
        self.assertEqual(str(update_error), "Invalid Request")



from enum import IntEnum

from mutenix.updates.constants import MAX_CHUNK_SIZE


class ChunkType(IntEnum):
    FILE_START = 1
    FILE_CHUNK = 2
    FILE_END = 3
    COMPLETE = 4
    FILE_DELETE = 5


class Chunk:
    def __init__(self, type_: int, id: int, package: int = 0, total_packages: int = 0):
        self.type_ = type_
        self.id = id
        self.package = package
        self.total_packages = total_packages
        self._acked = False
        self.content = b""

    def packet(self):  # pragma: no cover
        return (
            self._base_packet()
            + self.content
            + b"\0" * (MAX_CHUNK_SIZE - len(self.content))
        )

    def _base_packet(self):
        return (
            int(self.type_).to_bytes(2, "little")
            + self.id.to_bytes(2, "little")
            + self.total_packages.to_bytes(2, "little")
            + self.package.to_bytes(2, "little")
        )

    @property
    def acked(self):
        return self._acked


class FileChunk(Chunk):
    def __init__(self, id: int, package: int, total_packages: int, content: bytes):
        super().__init__(ChunkType.FILE_CHUNK, id, package, total_packages)
        self.content = content


class FileStart(Chunk):
    def __init__(
        self,
        id: int,
        package: int,
        total_packages: int,
        filename: str,
        filesize: int,
    ):
        super().__init__(ChunkType.FILE_START, id, package, total_packages)
        self.content = (
            bytes((len(filename),))
            + filename.encode("utf-8")
            + bytes((2,))
            + filesize.to_bytes(2, "little")
        )


class FileEnd(Chunk):
    def __init__(self, id: int):
        super().__init__(ChunkType.FILE_END, id)


class FileDelete(Chunk):
    def __init__(
        self,
        id: int,
        filename: str,
    ):
        super().__init__(ChunkType.FILE_DELETE, id)
        self.content = bytes((len(filename),)) + filename.encode("utf-8")


class Completed(Chunk):
    def __init__(self):
        super().__init__(ChunkType.COMPLETE, 0)



import pathlib

from aiohttp.test_utils import AioHTTPTestCase
from mutenix.web_server import WebServer


class TestWebServer(AioHTTPTestCase):
    async def get_application(self):
        return WebServer().app

    async def test_serve_image_success(self):
        image_name = "test_image.png"
        image_path = (
            pathlib.Path(__file__).parent.parent
            / "src"
            / "mutenix"
            / "assets"
            / image_name
        )
        image_path.touch()  # Create an empty file for testing

        request = await self.client.request("GET", f"/images/{image_name}")
        assert request.status == 200
        content = await request.read()
        assert content == image_path.read_bytes()

        image_path.unlink()  # Clean up the created file

    async def test_serve_image_not_found(self):
        request = await self.client.request("GET", "/images/non_existent_image.png")
        assert request.status == 404

    async def test_popup_success(self):
        request = await self.client.request("GET", "/popup")
        assert request.status == 200
        text = await request.text()
        assert "DOCTYPE html>" in text  # Adjust this assertion based on actual content

    async def test_popup_not_found(self):
        request = await self.client.request("GET", "/non_existent_popup")
        assert request.status == 404
